'''
Frames each have a .plot and .update method associated with them,
meaning that they can be included and animated inside a structured
Illustration.
'''

from .FrameBase import FrameBase
from .imshowFrame import imshowFrame
from .CameraFrame import CameraFrame, cameras
from .CCDFrame import CCDFrame, ccds
from .ZoomFrame import ZoomFrame
from .LocalZoomFrame import LocalZoomFrame
from .LocalStampFrame import LocalStampFrame
from .EmptyTimeseriesFrame import EmptyTimeseriesFrame



'''
Tools for setting up colors for display.
'''

from .imports import *
from matplotlib.colors import SymLogNorm, LogNorm

def cmap_norm_ticks(a, whatpercentiles=[1, 99], howmanysigmaarelinear=1.5, whatfractionislinear=0.15, vmax=None, vmin=None, cmap=None):
    '''
    Return a probably pretty-OK colormap, a color normalization,
    and suggested tick marks for a colorbar, based on an input array.

    Parameters
    ----------

    a : array
            The cmap and norm will be set on the basis of values in this array.
    '''

    if vmin is None:
        gonegative = (a <= 0).any()
    else:
        gonegative = vmin < 0

    # figure out a decent color scale
    if gonegative:
        # go diverging, if this is a difference that crosses zero
        if vmax is not None:
            vmin = -vmax
        else:
            vmin, vmax = np.nanpercentile(a, whatpercentiles)
            scale = np.maximum(np.abs(vmin), np.abs(vmax))
            vmin, vmax = -scale, scale
        span = np.log10(vmax)
        sigma = mad(a)

        norm = SymLogNorm(howmanysigmaarelinear * sigma,
                          linscale=span * whatfractionislinear,
                          vmin=vmin, vmax=vmax)
        if cmap is None:
            cmap = 'RdBu'
        ticks = [vmin, -howmanysigmaarelinear * sigma,
                 0, howmanysigmaarelinear * sigma, vmax]
    else:
        if vmax is None:
            vmax = np.nanpercentile(a, whatpercentiles[1])
        if vmin is None:
            vmin = np.nanpercentile(a, whatpercentiles[0])

        # go simple logarithmic, if this is all positive
        norm = LogNorm(vmin=vmin, vmax=vmax)
        if cmap is None:
            cmap = 'Blues'
        ticks = [vmin, vmin * np.sqrt(vmax / vmin), vmax]

    return cmap, norm, ticks



# still a WIP!
from .FourCameraIllustration import *


'''
class ZoomsIllustration(FourCameraIllustration):

    def __init__(self, stamps=[], **kw):

        # first, initialize the base FourCamera
        FourCameraIllustration.__init__(self, **kw)


        for i, s in enumerate(stamps):

            row, col = s.static['ROW_CENT'], s.static['COL_CENT']
            data_coords = col, row

            cam = s.static['CAM']
            self.speak(cam)
            reference_ax = self.frames[cam].ax
            reference_fi = reference_ax.figure

            display_coords = reference_ax.transData.transform((data_coords))
            figure_coords = reference_fi.transFigure.inverted().transform(display_coords)

            self.speak(cam, data_coords, display_coords, figure_coords)

            size = 0.2
            left, bottom = figure_coords
            width, height = size, size
            ax = plt.axes([left, bottom, width, height])

            key = 'STAMP{}'.format(i)
            self.frames[key] = imshowFrame(ax=ax, data=s, name=key)
'''



'''
Define some filename parsers than can pull details from filenames,
and convert them to dictionaries.
'''

from ..imports import *

def generic_filenameparser(filename):
    '''
    Parse a general file like, for example,
    "some-little-filename.fits"
    into a dictionary of important information.

    (This one is meant to be expanded upon elsewhere.)

    Parameters
    ----------
    filename : str
        The filename of a single image.

    Returns
    -------
    features : dict
        This dictionary of details contains the filename, (and you
        should add some other keywords too, if you want to expand
        upon this.)
    '''
    d = {}
    d['filename'] = os.path.basename(filename)
    return d

def explicit_filenameparser(filename):
    '''
    Parse a file that explicitly specifies ccd and camera
    like, for example, "cam1-ccd2-00001.fits"
    into a dictionary of important information.

    (This one is meant to be expanded upon elsewhere.)

    Parameters
    ----------
    filename : str
        The filename of a single image.

    Returns
    -------
    features : dict
        This dictionary of details contains the filename, camera, and CCD.
    '''
    d = {}
    d['filename'] = os.path.basename(filename)
    s = d['filename'].split('.fit')[0]
    d['camera'] = s[s.index('cam'):][3]
    d['ccd'] = s[s.index('ccd'):][3]
    return d

def tessqlp_filenameparser(filename):
    '''
    Parse a QLP-type (single-CCD) file like, for example,
    "tess2018220104526-00004907-1-crm-ffi-ccd1.fits"
    into a dictionary of important information.

    Parameters
    ----------
    filename : str
        The filename of a single FITS image.

    Returns
    -------
    features : dict
        This dictionary of details contains the filename, the cadence number,
        the camera number, whether or not crm was applied, the image type, and
        the CCD number.
    '''

    d = {}
    d['filename'] = os.path.basename(filename)
    s = d['filename'].split('.fit')[0]
    components = s.split('-')
    d['cadence'] = float(components[1])
    d['camera'] = int(components[2])
    d['crm'] = components[3]
    d['type'] = components[4]
    d['ccd'] = s[s.index('ccd'):][3]
    return d

def tessqlp_fullcamera_filenameparser(filename):
    '''
    Parse a QLP-type file like, for example,
    "tess2018220104526-00005279-4-crm-ffi_dehoc.fits"
    into a dictionary of important information.

    Parameters
    ----------
    filename : str
        The filename of a single FITS image.

    Returns
    -------
    features : dict
        This dictionary of details contains the filename, the cadence number,
        the camera number, whether or not crm was applied, and the image type.
    '''

    d = {}
    d['filename'] = os.path.basename(filename)
    s = d['filename'].split('.fit')[0]
    components = s.split('-')
    d['cadence'] = float(components[1])
    d['camera'] = int(components[2])
    d['crm'] = components[3]
    d['type'] = components[4]
    return d

def flexible_filenameparser(filename):
    '''
    Parse a generic FITS filename,
    trying to guess an appropriate method for it,
    and create a dictionary of important information.

    Parameters
    ----------
    filename : str
        The filename of a single FITS image.

    Returns
    -------
    features : dict
        This dictionary of details contains at least the filename, and
        hopefully some other stuff too.
    '''

    # filename parsers, in order of priority
    parsers = [ tessqlp_filenameparser,
                tessqlp_fullcamera_filenameparser,
                explicit_filenameparser,
                generic_filenameparser]

    # try each parser in turn, moving to the next if one complains
    for p in parsers:
        try:
            return p(filename)
        except:
            pass



from .imshowFrame import *
from astropy.nddata.utils import Cutout2D


class ZoomFrame(imshowFrame):
    frametype = 'Zoom'

    def __init__(self,  source=None,
                        position=(0, 0),
                        size=(10, 10),
                        name='zoom',
                        cmapkw=dict(),
                        plotingredients=[   'image',
                                            'colorbar'], #'arrows','title'],
                        **kwargs):
        '''
        Initialize a ZoomFrame that can display
        images taken from a zoomed subset of
        a `source` imshowFrame. This is kind
        of like a magnifying class you can
        drop down on an image.

        Parameters
        ----------

        source : imshowFrame
                A frame into which this one will be zooming.

        position : (tuple)
                The (x, y) = (col, row) position of the center of zoom.

        shape : (tuple)
                The (x, y) = (ncols, nrows) shape of the zoom.

        name : str
            A name to give this Frame.

        ax : matplotlib.axes.Axes instance
            All plotting will happen inside this ax.
            If set to None, the `self.ax attribute` will
            need to be set manually before plotting.

        plotingredients : list
            A list of keywords indicating features that will be
            plotted in this frame. It can be modified either now
            when initializing the frame, or any time before
            calling `i.plot()` from the illustration.

        cmapkw : dict
            Dictionary of keywords to feed into the cmap generation.
        '''

        imshowFrame.__init__(self,  name=name, plotingredients=plotingredients, **kwargs)

        self.source = source
        self.position = position

        # cutout needs (row, col), but we want to give (x,y)
        self.size = size[::-1]

        # set the aspect ratio (width/height) for this zoom frame
        self.aspectratio = float(self.size[1])/self.size[0] #width/height

        self.titlefordisplay = '{} | {}'.format(self.frametype, self.position)

        self.cmapkw = copy.copy(cmapkw) # why do I have to do this?

    def _get_times(self):
        '''
        Get the available times associated with this frame.
        '''
        return self.source._get_times()

    def _find_timestep(self, time):
        '''
        Get the timestep, by using the source frame.
        '''
        return self.source._find_timestep(time)

    def _get_image(self, time=None):
        '''
        Get the image at a given time (defaulting to the first time),
        by pulling it from the source frame.
        '''

        bigimage, actual_time = self.source._get_image(time)
        self.cutout = Cutout2D(bigimage, self.position,
                               self.size, mode='partial')
        cutoutimage = self.cutout.data
        return cutoutimage, actual_time

    def plot(self, *args, **kwargs):

        # do all the normal stuff for an imshow plot
        imshowFrame.plot(self, *args, **kwargs)

        # add a box to the source image (cutout must have ben created, if plot has happened)
        self.cutout.plot_on_original(ax=self.source.ax, clip_on=True)

    def update(self, time):
        '''
        Update this frame to a particular time (for use in animations).
        '''
        # update the data, if we need to
        timestep = self._find_timestep(time)
        image, actual_time = self._get_image(time)
        if image is None:
            return

        if timestep != self.currenttimestep:
            self.plotted['image'].set_data(image)
            if 'time' in self.plotted:
                self.plotted['time'].set_text(self._timestring(actual_time))
        self.currenttimestep = timestep

    def __repr__(self):
        '''
        Default string representation for this frame.
        '''
        return '<{} Frame | position={} | size={}>'.format(self.frametype, self.position, self.size)



from ..imports import *

from lightkurve.targetpixelfile import TargetPixelFile, KeplerTargetPixelFile, KeplerTargetPixelFileFactory, KeplerQualityFlags
from lightkurve import KeplerLightCurve, TessLightCurve
import datetime, warnings
from tqdm import tqdm


import os
TESSTPFDIR = os.path.abspath(os.path.dirname(__file__))


class EarlyTessLightCurve(KeplerLightCurve):
    def __init__(self, time, flux=None, flux_err=None,  tic_id=None, cam=None, spm=None, col_cent=None, row_cent=None, mission='TESS', **kwargs):
        self.tic_id = tic_id
        self.cam = cam
        self.spm = spm
        self.col_cent = col_cent
        self.row_cent = row_cent
        self.mission = mission
        KeplerLightCurve.__init__(self, time=time, flux=flux, flux_err=flux_err,  **kwargs)
        self.time_format = 'jd'
        self.time_scale = 'tdb'

    def plot(self, ax=None, xlabel='Time (days)', **kwargs):
        return KeplerLightCurve.plot(self, ax=None,  xlabel=xlabel, **kwargs)

    @property
    def cadence(self):
        return np.median(np.diff(self.time))*u.day


class EarlyTessTargetPixelFile(KeplerTargetPixelFile):
    """
    Defines a TargetPixelFile class for early (MIT) TESS data.
    Right now, this is based off '*_sparse_subarrays.fits' files,
    which may have been processed into Stamp objects. They can
    be loaded via the .from_stamp or .from_sparse_subarrays
    methods.

    Enables extraction of raw lightcurves and centroid positions,
    and some nice simple visualizations.
    """

    def __init__(self, *args, **kwargs):
        # KLUDGE, to make sure a flux gets defined, even for uncalibrated data
        KeplerTargetPixelFile.__init__(self, *args, **kwargs)
        if np.sum(np.isfinite(self.flux)) == 0:
            self.hdu[1].data['FLUX'] = self.raw_cnts

    @property
    def raw_cnts(self):
        return self.hdu[1].data['RAW_CNTS']

    @property
    def cadence(self):
        return np.median(np.diff(self.time))*u.day

    @property
    def background_mask(self):
        return self.hdu[-1].data == 2

    def filelabel(self, label=None):
        '''
        TESS-specific file label.
        '''
        s = 'cam{}_spm{}_tic{}_col{}_row{}_{:.0f}s'.format(self.cam, self.spm, self.tic_id, self.col_cent, self.row_cent, self.cadence.to('s').value)
        if label is not None:
            s = label + '_' + s
        return s

    def to_fits(self, output_fn=None, overwrite=False, directory='.', label=None, zip=True):
        """Writes the TPF to a FITS file on disk."""

        # set the output filename
        if output_fn is None:
            output_fn = os.path.join(directory, "{}-targ.fits".format(self.filelabel(label=label)))

        # skip writing the file, if a zipped one already exists
        if zip:
            if overwrite == False:
                zipped_fn = output_fn + '.gz'
                if os.path.exists(zipped_fn):
                    print('{} already exists! not overwriting!'.format(zipped_fn))
                    return

        # write the fits file
        try:
            self.hdu.writeto(output_fn, overwrite=overwrite, checksum=True)
        except OSError:
            print('(I think) {} already exists! not overwriting!'.format(output_fn))
        if zip:
            os.system('gzip -v {}'.format(output_fn))


    def to_lightcurve(self, aperture_mask='pipeline'):
        """Performs aperture photometry.

        Parameters
        ----------
        aperture_mask : array-like, 'pipeline', or 'all'
            A boolean array describing the aperture such that `False` means
            that the pixel will be masked out.
            If the string 'all' is passed, all pixels will be used.
            The default behaviour is to use the Kepler pipeline mask.

        Returns
        -------
        lc : KeplerLightCurve object
            Array containing the summed flux within the aperture for each
            cadence.
        """
        aperture_mask = self._parse_aperture_mask(aperture_mask)
        if aperture_mask.sum() == 0:
            log.warning('Warning: aperture mask contains zero pixels.')
        centroid_col, centroid_row = self.centroids(aperture_mask)
        keys = {'centroid_col': centroid_col,
                'centroid_row': centroid_row,
                'quality': self.quality,
                'mission': self.mission,
                'cadenceno': self.cadenceno,
                'cam':self.cam,
                'spm':self.spm,
                'col_cent':self.col_cent,
                'row_cent':self.row_cent,
                'tic_id':self.tic_id}

        return EarlyTessLightCurve( time=self.time,
                                    flux=np.nansum(self.flux[:, aperture_mask], axis=1),
                                    flux_err=np.nansum(self.flux_err[:, aperture_mask]**2, axis=1)**0.5,
                                    **keys)

    def differences(self, **kwargs):
        '''
        NEW!

        Calculate the differences between subsequent images,
        and return a new TPF.
        '''
        n_cadences, n_rows, n_cols = self.shape
        tic_id = self.tic_id


        print('making TPF from the differences of {}'.format(self))
        # create a factory to populate with pixels
        factory = EarlyTessTargetPixelFileFactory(n_cadences=n_cadences-1,
                                                   n_rows=n_rows,
                                                   n_cols=n_cols,
                                                   target_id=tic_id)

        # a stamp already exists, so we can add everything all at once
        factory.keywords = fits.Header(self.hdu[0].header)
        #factory.raw_cnts = np.diff(self.raw_cnts, axis=0)
        factory.flux = np.diff(self.flux, axis=0)
        #factory.flux_bkg = np.diff(self.flux_bkg, axis=0)

        # set some variables in the time dimension
        factory.time = 0.5*(self.time[1:] + self.time[:-1])
        factory.cadenceno = np.arange(len(factory.time))
        factory.quality = self.quality[1:] | self.quality[:-1]

        return factory.get_tpf(**kwargs)

    @staticmethod
    def from_archive(target, cadence='long', quarter=None, month=None,
                     campaign=None, radius=1., targetlimit=1, verbose=True, **kwargs):
        """
        There's no archive yet!
        """
        raise ValueError("There's no TESS archive yet!")

    def __repr__(self):
        return('EarlyTessTPF Object (TIC{})'.format(self.tic_id))

    def get_prf_model(self):
        """
        FIXME! This should turn into a very simple (Gaussian)
        and then slightly more complicated (database) PRF at some pointself.

        Returns an object of SimpleKeplerPRF initialized using the
        necessary metadata in the tpf object.

        Returns
        -------
        prf : instance of SimpleKeplerPRF
        """

        raise RuntimeWarning("PRF not yet made for TESS!")

        return SimpleKeplerPRF(channel=self.channel, shape=self.shape[1:],
                               column=self.column, row=self.row)

    @property
    def astropy_time(self):
        """Returns an AstroPy Time object for all good-quality cadences."""
        return Time(self.time, format='jd', scale='tdb')

    @property
    def tic_id(self):
        return self.header()['TIC_ID']

    @property
    def cam(self):
        return self.header()['CAM']

    @property
    def spm(self):
        return self.header()['SPM']

    @property
    def col_cent(self):
        return self.header()['COL_CENT']

    @property
    def row_cent(self):
        return self.header()['ROW_CENT']

    # add a thing to make reasonable WCS for each stamp -- can this be tied to TIC location + aperture definition?

    def plot(self, *args, **kwargs):
        """
        Plot a target pixel file at a given frame (index) or cadence number.

        Parameters
        ----------
        ax : matplotlib.axes._subplots.AxesSubplot
            A matplotlib axes object to plot into. If no axes is provided,
            a new one will be generated.
        frame : int
            Frame number. The default is 0, i.e. the first frame.
        cadenceno : int, optional
            Alternatively, a cadence number can be provided.
            This argument has priority over frame number.
        bkg : bool
            If True, background will be added to the pixel values.
        aperture_mask : ndarray
            Highlight pixels selected by aperture_mask.
        show_colorbar : bool
            Whether or not to show the colorbar
        mask_color : str
            Color to show the aperture mask
        style : str
            matplotlib.pyplot.style.context, default is 'fast'
        kwargs : dict
            Keywords arguments passed to `lightkurve.utils.plot_image`.

        Returns
        -------
        ax : matplotlib.axes._subplots.AxesSubplot
            The matplotlib axes object.
        """
        ax = KeplerTargetPixelFile.plot(self, *args, **kwargs)
        ax.set_title('TIC: {}'.format(self.tic_id))
        return ax

    @staticmethod
    def from_fits_images(images, position=None, size=(10, 10), extension=None,
                         target_id="unnamed-target", **kwargs):
        """
        work-in-progress -- (e.g. to extract single stamps from FFIs)

        Creates a new Target Pixel File from a set of images.

        This method is intended to make it easy to cut out targets from
        Kepler/K2 "superstamp" regions or TESS FFI images.

        Attributes
        ----------
        images : list of str, or list of fits.ImageHDU objects
            Sorted list of FITS filename paths or ImageHDU objects to get
            the data from.
        position : astropy.SkyCoord
            Position around which to cut out pixels.
        size : (int, int)
            Dimensions (cols, rows) to cut out around `position`.
        extension : int or str
            If `images` is a list of filenames, provide the extension number
            or name to use. Default: 0.
        target_id : int or str
            Unique identifier of the target to be recorded in the TPF.
        **kwargs : dict
            Extra arguments to be passed to the `KeplerTargetPixelFile` constructor.

        Returns
        -------
        tpf : KeplerTargetPixelFile
            A new Target Pixel File assembled from the images.
        """
        if extension is None:
            if isinstance(images[0], str) and images[0].endswith("ffic.fits"):
                extension = 1  # TESS FFIs have the image data in extension #1
            else:
                extension = 0  # Default is to use the primary HDU

        factory = EarlyTessTargetPixelFileFactory(n_cadences=len(images),
                                                   n_rows=size[0],
                                                   n_cols=size[1],
                                                   target_id=target_id)
        for idx, img in tqdm(enumerate(images), total=len(images)):
            if isinstance(img, fits.ImageHDU):
                hdu = img
            elif isinstance(img, fits.HDUList):
                hdu = img[extension]
            else:
                hdu = fits.open(img)[extension]
            if idx == 0:  # Get default keyword values from the first image
                factory.keywords = hdu.header
            if position is None:
                cutout = hdu
            else:
                cutout = Cutout2D(hdu.data, position, wcs=WCS(hdu.header),
                                  size=size, mode='partial')
            factory.add_cadence(frameno=idx, flux=cutout.data, header=hdu.header)
        return factory.get_tpf(**kwargs)

    @staticmethod
    def from_stamp(stamp, **kwargs):

        """
        Creates a new Target Pixel File from a stamp.

        Attributes
        ----------
        stamp : a stamp object
        **kwargs : dict
            Extra arguments to be passed to the `EarlyTessTargetPixelFile` constructor.

        Returns
        -------
        tpf : EarlyTessTargetPixelFile
            A new Target Pixel File assembled from the images.
        """

        n_cadences, n_rows, n_cols = stamp.photons.shape
        tic_id = stamp.static['TIC_ID']


        print('making TPF from {}'.format(stamp)
        )
        # create a factory to populate with pixels
        factory = EarlyTessTargetPixelFileFactory(n_cadences=n_cadences,
                                                   n_rows=n_rows,
                                                   n_cols=n_cols,
                                                   target_id=tic_id)

        # a stamp already exists, so we can add everything all at once
        factory.keywords = fits.Header(stamp.static)
        factory.raw_cnts = stamp.photons
        factory.flux = stamp.photons

        # set some variables in the time dimension

        gpstime = stamp.temporal['TIME']
        factory.time = Time(gpstime, format='gps', scale='tdb').jd
        factory.cadenceno = np.arange(len(gpstime))#stamp.temporal['CADENCE']
        factory.quality = stamp.temporal['QUAL_BIT']

        return factory.get_tpf(**kwargs)


    @staticmethod
    def from_sparse_subarrays(images,      # each image is a time-point
                              extension=1, # each extension is a stamp
                              target_id="unnamed-target",
                              **kwargs):
        """
        WIP?
        Creates a new Target Pixel File from a set of sparse images.
        This is meant to be mostly a tool to be run through all
        extensions of a sparse subarray, to create stamps that
        are saved as actual TPF files, indexed either by TIC or
        by camera coordinates.

        Attributes
        ----------
        images : list of str, or list of fits.ImageHDU objects
            Sorted list of FITS filename paths or ImageHDU objects to get
            the data from.
        extension : int or str
            Which extension (aka stamp) to extract?
        target_id : int or str
            Unique identifier of the target to be recorded in the TPF.
        **kwargs : dict
            Extra arguments to be passed to the `EarlyTessTargetPixelFile` constructor.

        Returns
        -------
        tpf : EarlyTessTargetPixelFile
            A new Target Pixel File assembled from the images.
        """

        #  we need an extension > 1 to get
        assert(extension > 0)

        # open the first one, to get the shape of the image
        first = fits.open(images[0])[extension]
        singleheader = first.header
        singleimage = first.data

        tic_id = singleheader['TIC_ID']
        '''

        size = singleimage.shape[0]
        array = np.empty((N, data.shape[0], data.shape[1]))
        oned = {}
        keys = ['INT_TIME', 'QUAL_BIT', 'TIME', 'CADENCE']
        for k in keys:
            oned[k] = np.empty(N)

        tic = header['TIC_ID']
        location = header['COL_CENT'], header['ROW_CENT']
        '''


        print('making TPF from ext={} of {} files'.format(extension, len(images)))
        # create a factory to populate with pixels
        factory = EarlyTessTargetPixelFileFactory(n_cadences=len(images),
                                                   n_rows=singleimage.shape[0],
                                                   n_cols=singleimage.shape[1],
                                                   target_id=tic_id)

        for idx, img in enumerate(tqdm(images)):

            # allow us to send hdu lists or filenames
            if isinstance(img, fits.HDUList):
                hdus = img
            else:
                hdus = fits.open(img)

            stamphdu = hdus[extension]
            framehdu = hdus[0]


            if idx == 0:  # Get default keyword values from the first image
                factory.keywords = framehdu.header
                for k in stamphdu.header.keys():
                    factory.keywords[k] = stamphdu.header[k]

            # original times are in GPS seconds
            gpstime = framehdu.header['TIME']

            # let's convert them to jd
            framehdu.header['TIME'] = Time(gpstime, format='gps', scale='tdb').jd

            # add this cadence to the TOP
            factory.add_cadence(frameno=idx, raw_cnts=stamphdu.data, flux=stamphdu.data, header=framehdu.header)

        return factory.get_tpf(**kwargs)

    @property
    def keplerid(self):
        return None


class EarlyTessTargetPixelFileFactory(KeplerTargetPixelFileFactory):
    def add_cadence(self, frameno, raw_cnts=None, flux=None, flux_err=None,
                    flux_bkg=None, flux_bkg_err=None, cosmic_rays=None,
                    header={}):
        """Populate the data for a single cadence."""
        # 2D-data
        for col in ['raw_cnts', 'flux', 'flux_err', 'flux_bkg',
                    'flux_bkg_err', 'cosmic_rays']:
            if locals()[col] is not None:
                vars(self)[col][frameno] = locals()[col]

        # 1D-data
        if 'TIME' in header:
            # FIXME -- currently GPS time (is it start/mid/end of exposure?)
            self.time[frameno] = header['TIME']

        if 'TIMECORR' in header:
            self.timecorr[frameno] = header['TIMECORR']
        if 'CADENCE' in header:
            self.cadenceno[frameno] = header['CADENCE']
        if 'QUALITY' in header:
            self.quality[frameno] = header['QUAL_BIT']

        # FIXME -- figure out how to deal with positions?
        if 'POSCORR1' in header:
            self.pos_corr1[frameno] = 0# header['POSCORR1']
        if 'POSCORR2' in header:
            self.pos_corr2[frameno] = 0# header['POSCORR2']


    def get_tpf(self, **kwargs):
        """Returns a KeplerTargetPixelFile object."""

        # FIXME -- storing the factory is just for debugging
        tpf = EarlyTessTargetPixelFile(self._hdulist(), **kwargs)
        tpf.factory = self
        return tpf

    def _header_template(self, extension):
        """
        Returns a template `fits.Header` object for a given extension,
        specifically tweaked for Early TESS objects.
        """
        template_fn = os.path.join(TESSTPFDIR, "data",
                                   "tpf-ext{}-header.txt".format(extension))
        return fits.Header.fromtextfile(template_fn)


    ### FIXME! You need to understand the header templates!
    def _make_primary_hdu(self, hdu0_keywords={}, keywords={}, **kw):
        """Returns the primary extension (#0)."""
        hdu = fits.PrimaryHDU()
        # Copy the default keywords from a template file from the MAST archive
        tmpl = self._header_template(0)
        for kw in tmpl:
            hdu.header[kw] = (tmpl[kw], tmpl.comments[kw])
        # Override the defaults where necessary
        hdu.header['ORIGIN'] = "Unofficial data product"
        hdu.header['DATE'] = datetime.datetime.now().strftime("%Y-%m-%d")
        hdu.header['CREATOR'] = "lightkurve"
        hdu.header['OBJECT'] = self.target_id
        hdu.header['TIC_ID'] = self.target_id

        # Empty a bunch of keywords rather than having incorrect info
        for kw in ["PROCVER", "FILEVER", "CHANNEL", "MODULE", "OUTPUT",
                   "TIMVERSN", "CAMPAIGN", "DATA_REL", "TTABLEID",
                   "RA_OBJ", "DEC_OBJ"]:
            hdu.header[kw] = ""


        # FIXME -- (probably a kludge, maybe we need these only in ext=1)
        template = self._header_template(0)
        for kw in template:
            if kw not in ['XTENSION', 'NAXIS1', 'NAXIS2', 'CHECKSUM', 'BITPIX']:
                try:
                    hdu.header[kw] = (self.keywords[kw],
                                      self.keywords.comments[kw])
                except KeyError:
                    #print("Couldn't add [{}]".format(kw))
                    hdu.header[kw] = (template[kw],
                                      template.comments[kw])

        return hdu

    def _make_target_extension(self, **kw):
        """Create the 'TARGETTABLES' extension (i.e. extension #1)."""
        # Turn the data arrays into fits columns and initialize the HDU
        coldim = '({},{})'.format(self.n_cols, self.n_rows)
        eformat = '{}E'.format(self.n_rows * self.n_cols)
        jformat = '{}J'.format(self.n_rows * self.n_cols)
        cols = []

        # FIXME -- may need to modify this?
        cols.append(fits.Column(name='TIME', format='D', unit='spacecraft seconds',
                                array=self.time))
        cols.append(fits.Column(name='TIMECORR', format='E', unit='D',
                                array=self.timecorr))
        cols.append(fits.Column(name='CADENCENO', format='J',
                                array=self.cadenceno))
        cols.append(fits.Column(name='RAW_CNTS', format=jformat,
                                unit='count', dim=coldim,
                                array=self.raw_cnts))

        # FIXME -- at some point, switch these to being calibrated
        cols.append(fits.Column(name='FLUX', format=eformat,
                                unit='count', dim=coldim,
                                array=self.flux))
        cols.append(fits.Column(name='FLUX_ERR', format=eformat,
                                unit='count', dim=coldim,
                                array=self.flux_err))
        cols.append(fits.Column(name='FLUX_BKG', format=eformat,
                                unit='count', dim=coldim,
                                array=self.flux_bkg))
        cols.append(fits.Column(name='FLUX_BKG_ERR', format=eformat,
                                unit='count', dim=coldim,
                                array=self.flux_bkg_err))
        cols.append(fits.Column(name='COSMIC_RAYS', format=eformat,
                                unit='count', dim=coldim,
                                array=self.cosmic_rays))
        cols.append(fits.Column(name='QUALITY', format='J',
                                array=self.quality))
        cols.append(fits.Column(name='POS_CORR1', format='E', unit='pixels',
                                array=self.pos_corr1))
        cols.append(fits.Column(name='POS_CORR2', format='E', unit='pixels',
                                array=self.pos_corr2))
        coldefs = fits.ColDefs(cols)
        hdu = fits.BinTableHDU.from_columns(coldefs)

        # Set the header with defaults
        template = self._header_template(1)
        for kw in template:
            if kw not in ['XTENSION', 'NAXIS1', 'NAXIS2', 'CHECKSUM', 'BITPIX']:
                try:
                    hdu.header[kw] = (self.keywords[kw],
                                      self.keywords.comments[kw])
                except KeyError:
                    #print("Couldn't add [{}]".format(kw))
                    hdu.header[kw] = (template[kw],
                                      template.comments[kw])

        # Override the defaults where necessary
        hdu.header['EXTNAME'] = 'TARGETTABLES'
        hdu.header['OBJECT'] = self.target_id
        hdu.header['TIC_ID'] = self.target_id
        for n in [5, 6, 7, 8, 9]:
            hdu.header["TFORM{}".format(n)] = eformat
            hdu.header["TDIM{}".format(n)] = coldim
        hdu.header['TFORM4'] = jformat
        hdu.header['TDIM4'] = coldim

        return hdu



from .IllustrationBase import *
from .CameraIllustration import *
from ..frames import CameraFrame, cameras
from ..frames import CCDFrame, ccds
__all__ = ['CameraOfCCDsIllustration']

class CameraOfCCDsIllustration(CameraIllustration):
    '''
    An illustration for displaying a single Camera (unrotated, untransformed).
    '''
    illustrationtype = 'CameraOfCCDsIllustration'

    def __init__(self,
                 ccd1=[], ccd2=[], ccd3=[], ccd4=[],
                 data=[],
                 orientation='horizontal',
                 sizeofcamera=4,
                 subplot_spec=None,
                 camera='camera',
                 sharecolorbar=True,
                 **framekw):
        '''
        Parameters
        ----------
        ccd1, ccd2, ccd3, ccd4
            Each should be able to initialize a sequence with
            the make_image_sequence() helper function.

        data
            Will be ignored.

        orientation : str
            'horizontal' = camera 2 is right of camera 1
            'vertical'  = camera 2 is below camera 1

        sizeofcamera : float
            What's the size, in inches, to display a single camera?
            (Each CCD will be half this.)

        subplot_spec : matplotlib.gridspec.SubplotSpec
            Should this illustration go into an
            existing SubplotSpec somewhere? (Useful for
            making subillustrations of other illustrations.)

        camera : str
            We need a CameraFrame associated with this
            illustration, to be able to handle transformations
            from Camera coordinates to display coordinates.
            Specify it here via 'camera' (default, no transformation)
            or 'cam1', 'cam2', 'cam3', 'cam4'.

        **framekw passed to frame
        '''


        # set up the basic geometry of the main axes
        rows, cols = 2, 2

        # what's the orientation of this illustration?
        self.orientation = orientation

        # create the basic illustration, possibly populating an existing subplot_spec
        IllustrationBase.__init__(self, rows, cols,
                                  figkw=dict(
                                      figsize=( sizeofcamera,
                                                sizeofcamera*1.25)),
                                  hspace=0.02, wspace=0.02,
                                  left=0.05, right=0.95,
                                  bottom=0.1, top=0.9,
                                  subplot_spec=subplot_spec,
                                  sharecolorbar=sharecolorbar)

        # FIXME -- make some more precise adjustments to this
        # to ensure that the spacing between CCDs doesn't get wonky
        # We could get pretty close simply by chosing the figsize
        # carefully whenever we make one, but it'd be good to have
        # much more self-contained control over the layout within
        # this camera.

        # create a hidden frame, which we'll use for transformations
        self._cameraframe = cameras[camera](illustration=self)

        assert(self._cameraframe.name == camera)


        if self.orientation == 'horizontal':
            # start the CCDs in the orientation of
            # 2 1
            # 3 4
            loc = np.array([[0,1], [0,0], [1,0], [1,1]])

            if self._cameraframe.transpose:
                # swap CCDs diagonally
                loc = loc[[0, 3, 2, 1]]

            if self._cameraframe.flipx:
                # swap CCDs along rows
                loc = loc[[1, 0, 3, 2]]

            if self._cameraframe.flipy:
                # swap CCDs along cols
                loc = loc[[3, 2, 1, 0]]

        # populate the axes on the main camera grid
        ax = {'ccd{}'.format(i+1):plt.subplot(self.grid[loc[i][0], loc[i][1]]) for i in range(4)}

        # loop through, create a frame for each CCD
        for k in ax.keys():
            self.frames[k] = ccds[k](illustration=self,
                                     ax=ax[k],
                                     data=make_image_sequence(locals()[k]),
                                     camera=self._cameraframe,
                                     **framekw)

            # by linking this CCD to a camera, we can use its transformations
            assert(self.frames[k].camera == self._cameraframe)

        self._condense_timelabels()



'''
Tools for adding zoom frames to an illustration.
'''

from .imports import *
from .illustrations import *
from .frames import ZoomFrame, LocalZoomFrame, LocalStampFrame

__all__ = ['add_grid_of_zooms', 'add_zoom', 'add_stamp']


def add_zoom(illustration, position, size=(25, 25), zoom=5, frame='camera'):
    '''
    Add a LocalZoomFrame to an illustration, which will appear
    as a magnified window centered at roughly its position on
    the original reference image.

    Parameters
    ----------
    illustration : an Illustration
        This is the overall illustration to which
        we want to be adding a local zoom.
    position : tuple
        (x,y) position of where the zoom should be centered
    size : tuple
        (nrows, ncols) shape of the region to zoom
    zoom : float
        By what factor do we magnify, relative to the original pixels
    frame : str
        Must exist as a key in the existing illustration.
    '''

    # FIXME - one option would be to make this dependent only on
    # the frame object itself, instead of using a frame keyword to
    # kludgily select a frame out of the illustration.

    # to which frame do we add this?
    reference_frame = illustration.frames[frame]

    # define a key for this frame
    key = 'zoom-{}-({},{})'.format(frame, position[0], position[1])
    illustration.frames[key] = LocalZoomFrame(illustration=illustration,
                                              ax=None,
                                              source=reference_frame,
                                              position=position,
                                              size=size,
                                              zoom=zoom)

    return illustration.frames[key]


def add_grid_of_zooms(illustration, N=4, buffer=100, cameras=[1, 2, 3, 4], zoom=10, size=10, **kw):
    '''
    Add a grid of local zooms to an an illustration containing four cameras.
    '''

    # FIXME - this a little specific of a use case right now. It might be
    # a better, more general tool if it simply added a grid of frames
    # to any particular frame object, instead of trying to pull the
    # frames out of the illustration. (If so, keep this wrapper, as a quick)
    # tool for adding a grid of zooms to a FourCameraIillustration (although)
    # that won't work for a CCD-based one.

    for c in cameras:
        f = illustration.frames['cam{}'.format(c)]

        buffer = size * zoom
        for row in np.linspace(f.ymin + buffer, f.ymax - buffer, N):
            for col in np.linspace(f.xmin + buffer, f.xmax - buffer, N):
                add_zoom(illustration, position=(col, row),
                         frame='cam{}'.format(c), size=size, zoom=zoom, **kw)


def add_stamp(illustration, stamp, zoom=5, camera='camera'):
    '''
    Add a Stamp to an illustration, at roughly its position.

    Parameters
    ----------
    illustration : an Illustration
        This is the overall illustration to which
        we want to be adding a local zoom.
    position : tuple
        (x,y) position of where the zoom should be centered
    size : tuple
        (nrows, ncols) shape of the region to zoom
    zoom : float
        By what factor do we magnify, relative to the original pixels
    camera : str
        Must exist as a key in the illustration.frames
    '''

    # to which frame do we add this?
    reference_frame = illustration.frames[camera]
    position = stamp.static['COL_CENT'], stamp.static['ROW_CENT']

    # define a key for this frame
    key = 'stamp-{}-({},{})'.format(camera, position[0], position[1])
    illustration.frames[key] = LocalStampFrame(illustration=illustration,
                                               ax=None,
                                               data=stamp,
                                               source=reference_frame,
                                               position=position,
                                               zoom=zoom)

    return illustration.frames[key]



'''
Tools to stack images, using a variety of algorithms
(this duplicates many of the algorithms from Strategies,
so that they can be better optimized for 2D images).
'''

from ..imports import *

__all__ = ['Central', 'Sum']

class Stacker(Talker):
    '''Stack a cube of images, using some filter.'''

    def __init__(self, **kwargs):
        # decide whether or not this Stacker is chatty
        Talker.__init__(self, **kwargs)

    def __repr__(self):
        return "<{}>".format(self.name)

    def average1d(self, array, nsubexposures=1):
        subexposures,  = array.shape
        exposures = subexposures//nsubexposures

        trim = exposures*nsubexposures
        trimmed = array[:trim]

        # reshape into something more convenient for summing
        splitintosubexposures = trimmed.reshape(exposures, nsubexposures)

        return np.sum(splitintosubexposures, 1)/nsubexposures

class Sum(Stacker):
    '''
    Binning with Sum = simply sum along the time axis.
    '''
    def __init__(self, **kwargs):
        self.name = 'Sum'

    def __call__(self, array, nsubexposures=1):
        '''
        Parameters:
        -----------
        array : a (xsize, ysize, time) array
            Probably the photons detected.
        nsubexposures : int
            How subexposures are being stacked together?
            (this will be ignored for the straight sum)
        '''

        subexposures, xpixels, ypixels = array.shape
        exposures = subexposures//nsubexposures

        trim = exposures*nsubexposures
        trimmed = array[:trim, :, :]

        # reshape into something more convenient for summing
        splitintosubexposures = trimmed.reshape(exposures, nsubexposures, xpixels, ypixels)

        return np.sum(splitintosubexposures, 1)


class Central(Stacker):
    '''
    Binning with TruncatedMean = break into subsets,
    reject the highest and lowest points from each and
    take the mean of the rest, and sum these truncated means.

    (rescales by N/M to achieve the same values as a sum)
    '''
    def __init__(self, n=10, m=None, **kwargs):
        Stacker.__init__(self)
        self.n = n
        if m is None:
            self.m = self.n-2
        else:
            self.m = m
            assert(((self.n - self.m) % 2) == 0)
        self.name = "Central {m} out of {n}".format(m = self.m, n=self.n)

    def __call__(self, array, nsubexposures):
        '''
        Parameters:
        -----------
        array : a (xsize, ysize, time) array
            Probably the photons detected.
        nsubexposures : int
            How subexposures are being stacked together?
        '''

        # this can handle only min-max rejection
        assert(self.n - self.m == 2)

        # figure out the original shape
        subexposures, xpixels, ypixels = array.shape
        exposures = subexposures//nsubexposures
        trim = exposures*nsubexposures
        trimmed = array[:trim, :, :]

        # reshape into something more convenient for summing
        splitintosubexposures = trimmed.reshape(exposures, nsubexposures, xpixels, ypixels)
        splitintochunks = splitintosubexposures.reshape( exposures, int(nsubexposures//self.n), self.n, xpixels, ypixels)

        # calculate the sum of the truncated means (and recalibrate to the original scale!!!)
        sum = np.sum(splitintochunks, 2)
        min = np.min(splitintochunks, 2)
        max = np.max(splitintochunks, 2)

        # rescale back to the original scale
        photons = np.sum(sum - max - min, 1)*self.n/self.m

        return photons.squeeze()



'''
`illumination` is a set of tools for display TESS pixels in Python and
matplotlib. It can be used for automated visualizations and animations,
so you can pull out your loupe and flyswatter and start looking at stars.
'''

# make sure we keep track of the version
from .version import __version__

# import some of the basics to be generally available
from .illustrations import *
from .zoom import *
from .sequences import *
from .wrappers import *
from .cartoons import *



from .IllustrationBase import *

__all__ = ['StampsIllustration']


class StampsIllustration(IllustrationBase):
    illustrationtype = 'Stamps'

    def __init__(self, cubes=[], names=[], aspectratio=np.inf, **kwargs):

        each = 3.2
        N = len(cubes)
        cols = np.minimum(np.ceil(np.sqrt(N) * aspectratio), N).astype(np.int)
        rows = np.maximum(np.ceil(N / cols), 1).astype(np.int)
        hspace, wspace = 0.15, 0.08
        left, right = 0.05, 0.95
        bottom, top = 0.15, 0.82
        wsize = each * cols * (1 + (cols - 1) * wspace) / (right - left)
        hsize = each * rows * (1 + (rows - 1) * hspace) / (top - bottom)

        IllustrationBase.__init__(self, rows, cols,
                                  figkw=dict(figsize=(wsize, hsize)),
                                  hspace=hspace, wspace=wspace,
                                  left=left, right=right,
                                  bottom=bottom, top=top, **kwargs)

        for i in range(rows):
            for j in range(cols):

                if len(cubes) == 0:
                    break

                ax = plt.subplot(self.grid[i, j])
                c = cubes.pop(0)
                try:
                    n = names[i * cols + j]
                except IndexError:
                    n = (i, j)

                self.frames[n] = imshowFrame(illustration=self, ax=ax, data=c)



'''
This constructs a stamp, as a Cube, from a directory of
sparse_subaray files.
'''

from ..imports import *
from .cubes import Cube
from astropy.time import Time
from lightkurve.targetpixelfile import TargetPixelFile

class Stamp(Cube):
	"""
	TESS postage stamp. Similar to a lightweight version of lightkurve.
	"""

	def __init__(self, path=None, extension=1, limit=None,
				 photons=None, spatial={}, static={}, temporal={}, **kw):
		'''
		Basically a light-weight version of a TPF.

		Parameters
		----------
		path : str, list of strings, list of HDULists, a TPF object
			A filename for a '.npy' file.
			A list of FITS sparse_subarray files.
			A search path of FITS sparse_subarray files (e.g. containing *.fits)
		extension : int
			Which stamp to pull from the subarray?
		limit : int
			How many files (in the search path) should we stop at?
		cube_kw : dict
			The inputs for a Cube.
		'''

		# figure out how to load the stamp
		if photons is not None:
			Cube.__init__(self, photons=photons, spatial=spatial, static=static, temporal=temporal)
		else:
			if isinstance(path, TargetPixelFile):
				self._fromTPF(path)
			elif '.npy' in path:
				self.load(path)
			else:
				if type(path) == list:
					filenames = path
				elif '*' in path:
					filenames = glob.glob(path)[:limit]
				else:
					raise ValueError("{} can't be used to make a stamp.".format(path))
				self._fromSparseSubarrays(filenames, extension)

		self.tic_id = self.static['TIC_ID']
		self.cadence = self.static['INT_TIME']
		self.spm = self.static['SPM']
		self.cam = self.static['CAM']

		self.identifier= 'TIC{TIC_ID} (CAM{CAM} | {ROW_CENT}, {COL_CENT})'.format(**self.static)
		self.label = 'SPM{SPM} | {INT_TIME}s'.format(**self.static)

		# default to some particular view
		self.consider('counts')


	def _fromSparseSubarrays(self, filenames, extension=1):
		'''
		Create a stamp object, from one extension of a group
		of FITS files make by faus' sparse_subarray code.

		This can be slow, if you need to keep reading the entire
		subarray file every time you want to load each extension.

		For faster,
		'''

		# open the first one, to get some basic info
		hdus = fits.open(filenames[0])
		frame = hdus[0]
		star = hdus[extension]
		static = {} #star.header

		# make empty temporal arrays
		N = len(filenames)

		for key in ['SPM', 'CAM', 'INT_TIME']:
			static[key] = frame.header[key]
		for key in ['TIC_ID', 'COL_CENT', 'ROW_CENT']:
			static[key] = star.header[key]



		temporal = {}
		for key in ['QUAL_BIT', 'TIME', 'CADENCE']:
			temporal[key] = np.empty(N)
			temporal[key][0] = frame.header[key]
		int_time, spm, cam = static['INT_TIME'], static['SPM'], static['CAM']
		# nothing (yet) that acts as a spatial image
		spatial = {}

		# KLUDGE, to convert ccd1 and ccd2 to camaer coords
		#flip = static['COL_CENT'] < 4272/2
		#flip = static['ROW_CENT'] < 4156/2

		# make empty photon arrays
		#if flip:
		#	data = star.data.T
		#else:
		data = star.data
		photons = np.empty((N, data.shape[0], data.shape[1]))

		# populate each time point
		for i, f in enumerate(filenames):
			#print(i, f)

			# the 0th extension contains time-dependent info
			hdu = fits.open(f)
			h0, d0 = hdu[0].header, hdu[0].data

			# make sure we're still on the right one
			assert(h0['INT_TIME'] == int_time)
			assert(h0['SPM'] == spm)
			assert(h0['CAM'] == cam)
			for k in temporal.keys():
				temporal[k][i] = h0[k]

			# the 1st extension contains the data for this star
			h, d = hdu[extension].header, hdu[extension].data
			#if flip:
			#	photons[i,:,:] = d.T[::-1, ::-1]
			#else:
			photons[i,:,:] = d

		self.__init__(self, photons=photons, temporal=temporal, spatial=spatial, static=static)
		self.speak('populated {}'.format(self))
		#if flip:
		#	self.speak('applied a KLUDGE to CCD1 + CCD2 (COL_CENT<2136?)')

	def _fromTPF(self, tpf):
		'''
		Make a Stamp from a TPF file.
		'''

		# open the first one, to get some basic info
		static = {}
		for k in ['SPM', 'CAM', 'INT_TIME'] + ['TIC_ID', 'COL_CENT', 'ROW_CENT'] + ['QUAL_BIT', 'TIME', 'CADENCE']:
			try:
				static[k] = tpf.hdu[0].header[k]
			except KeyError:
				pass
		# make empty temporal arrays
		N = len(tpf.time)


		temporal = {}
		temporal['QUAL_BIT'] = tpf.quality
		temporal['TIME'] = tpf.astropy_time.gps
		temporal['CADENCE'] = tpf.cadenceno

		# nothing (yet) that acts as a spatial image
		spatial = {}

		# KLUDGE, to convert ccd1 and ccd2 to camaer coords
		#flip = static['COL_CENT'] < 4272/2
		#flip = static['ROW_CENT'] < 4156/2

		# make empty photon arrays
		#if flip:
		#	data = star.data.T
		#else:
		photons = tpf.flux

		self.__init__(self, photons=photons, temporal=temporal, spatial=spatial, static=static)
		self.speak('populated {}'.format(self))
		#if flip:
		#	self.speak('applied a KLUDGE to CCD1 + CCD2 (COL_CENT<2136?)')


	def calculate_differences(self):

		difference = copy.deepcopy(self)
		for k in self.temporal.keys():
			difference.temporal[k] = 0.5*(self.temporal[1:] + self.temporal[:-1])

		difference.photons = np.diff(self.photons, )

	def filename(self, directory='.'):
		'''The base filename for this stamp.'''

		#return os.path.join(directory, 'tic{TIC_ID}_{ROW_CENT}-{COL_CENT}_cam{CAM}_spm{SPM}_{INT_TIME}s.npy'.format(**self.static))
		return os.path.join(directory, 'cam{CAM}_spm{SPM}_tic{TIC_ID}_col{COL_CENT}_row{ROW_CENT}_{INT_TIME}s.npy'.format(**self.static))

	def load(self, filename):
		'''
		Load from a .npy file.
		'''

		# make sure we're dealing with a npy saved file
		assert('.npy' in filename)
		loaded = np.load(filename, encoding='latin1')[()]
		self.__init__(self, **loaded)
		self.speak('loaded from {}'.format(filename))

	def save(self, filename):
		'''
		Save to a .npy file.
		'''

		# make sure we're dealing with a npy saved file
		assert('.npy' in filename)

		# populate a dictionary to save
		tosave = {}
		for k in self._savable:
			tosave[k] = vars(self)[k]
		np.save(filename, tosave)
		self.speak('saved to {} at {}'.format(filename, Time.now().iso))



from .imports import *

def get_writer(filename, fps=30, **kw):
    '''
    Try to get an appropriate animation writer,
    given the filename provided.

    Parameters
    ----------

    filename : str
        The output filename string for the animation.

    fps : float
        Frames/second.

    kw : dict
        All other keywords will be passed to the initialization
        of the animation writer.
    '''
    if '.mp4' in filename:
        try:
            writer = ani.writers['ffmpeg'](fps=fps, **kw)
        except (RuntimeError, KeyError):
            raise RuntimeError('This computer seems unable to ffmpeg.')
    else:
        try:
            writer = ani.writers['pillow'](fps=fps, **kw)
        except (RuntimeError, KeyError):
            writer = ani.writers['imagemagick'](fps=fps, **kw)
            raise RuntimeError('This computer seem unable to animate?')
    return writer


def guess_time_format(t, default='jd'):
    '''
    For a given array of times,
    make a guess about its time format.

    Parameters
    ----------
    t : array, float
            A time, in any format. This will try to guess the format,
            assuming we're dealing with data generally in the 2000s.

    default : str
            The default format, if no actual choice can be made.

    Returns
    -------
    format : str
            A time format string appropriate for astropy times.
    '''

    if t is []:
        return default

    # make some string guesses
    if isinstance(t[0], str):
        if 'T' in t[0]:
            return 'isot'
        else:
            return 'iso'

    ranges = dict( gps=[0.1e9, 2e9],  # valid between 1983-03-08 09:46:59.000 and 2043-05-23 03:33:39.000
                   # valid between 1858-11-16 12:00:00.000 and 3501-08-15 12:00:00.000
                   jd=[2.4e6, 3e6],
                   mjd=[4e4, 8e4])  # valid between 1968-05-24 00:00:00.000 and 2077-11-28 00:00:00.000

    for k in ranges.keys():
        if np.min(t) >= ranges[k][0] and np.max(t) <= ranges[k][1]:
            return k

    return default



'''
Define a generic sequence of images.
'''
from .Sequence import *
from ..cartoons import create_test_times

class Image_Sequence(Sequence):
    def __init__(self, initial, name='cube', time=None, temporal={}, spatial={}, **kwargs):
        '''
        Initialize a Sequence from a Cube.
        (a Cube is a custom, simplified, set of images)

        Parameters
        ----------
        initial : (many possible types)
            - a (ysize x = xsize)-shaped array
            - a (ntimes x ysize x = xsize)-shaped array

        time : array, None
            An array of times for the sequence.

        '''


        # create a sequence
        Sequence.__init__(self, name=name)

        # make sure we're dealing with an array
        array = np.atleast_2d(initial)

        # handle a single image as a 1-element array
        if len(array.shape) == 2:
            array = array[np.newaxis, :, :]
        elif len(array.shape) != 3:
            raise RuntimeError("The inputs to Image_Sequence seem to be the wrong shape.")

        # pull out the shape of the array
        N, ysize, xsize = array.shape
        self.images = array

        self.temporal = temporal
        self.spatial = spatial

        if time is None:
            self.time = Time(np.arange(N), format='gps', scale='tdb')
            self._timeisfake = True
        else:
            self._timeisfake = False
            assert(len(time) == N)
            if isinstance(time, Time):
                self.time = time
            else:
                self.time = Time(time, format=guess_time_format(time), scale=timescale)

    def __getitem__(self, timestep):
        '''
        Return the image data for a given timestep.

        This function is called when you say `sequence[timestep]`.

        Parameters
        ----------
        timestep : int
            A timestep index (which element in the sequence do you want?)
        '''
        if timestep is None:
            return None
        else:
            return self.images[timestep, :, :]

    @property
    def shape(self):
        '''
        Returns
        -------
        s : tuple
            The shape of the image stack (ntimes x nrows x ncols)
        '''
        d = self[0]
        return (self.N, d.shape[0], d.shape[1])

    def _gather_3d(self):
        '''
        Gather a 3D cube of images.

        This will generally work, but it
        may be faster if you already have
        access to the full 3D array stored
        in memory (as in a Stamp or TPF).

        Returns
        -------
        s : array
        The image stack, with shape (ntimes x nrows x ncols)
        '''

        s = np.zeros(self.shape)
        self.speak('gathering the sequence cube, with shape {}'.format(self.shape))
        for i in range(self.N):
            self.speak(' loaded frame {}/{}'.format(i+1, self.N), progress=True)
            s[i, :, :] = self[i]
        return s

    def median(self):
        '''
        Calculate the median image.

        Returns
        -------
        median : 2D image
            The median of the image sequence.
        '''

        try:
            self.spatial['median']
        except KeyError:
            self.speak('creating a median image for {}'.format(self))
            s = self._gather_3d()
            self.spatial['median'] = np.nanmedian(s, axis=0)
        return self.spatial['median']

    def sum(self):
        '''
        Calculate the sum of all the images.

        Returns
        -------
        sum : 2D image
            The median of the image sequence.
        '''

        s = self._gather_3d()
        return np.nansum(s, axis=0)

    def mean(self):
        '''
        Calculate the sum of all the images.
        It works in an inline fashion, so you
        don't need to load the entire image
        cube into memory (that might get big!)

        Returns
        -------
        mean : 2D image
            The median of the image sequence.
        '''

        try:
            self.spatial['mean']
        except KeyError:
            self.speak('creating a mean image for {}'.format(self))

            # calculate the mean in a running fashion (less memory)
            total = np.zeros_like(self[0])
            for i in range(self.N):
                self.speak(' included frame {}/{} in mean'.format(i+1, self.N), progress=True)
                thisimage = self[i]
                ok = np.isfinite(thisimage)
                total[ok] += thisimage[ok]

            self.spatial['mean'] = total/self.N

        return self.spatial['mean']

    def __repr__(self):
        '''
        How should this sequence be represented, by default, as a string.
        '''
        return '<{} of {} images of shape {}>'.format(self.nametag, self.shape[0], self.shape[1:] )



'''
Define a sequence of images from a TPF.
'''

from .Image_Sequence import *

class TPF_Sequence(Image_Sequence):
    def __init__(self, initial, name='TPF', **kwargs):
        '''
        Initialize a Sequence from a Target Pixel File
        (see lightkurve).


        Parameters
        ----------
        initial :
                -a TargetPixelFile object
                -a filename of a TargetPixelFile

        **kwargs : dict
                Keyword arguments are stored in self.keywords.
        '''

        # make sure we have a TPF as imput
        if type(initial) == str:
            tpf = EarlyTessTargetPixelFile.from_fits(initial)
        else:
            tpf = initial

        # create a sequence out of that stamp
        Sequence.__init__(self, name=name)

        # set up the basic sequence
        self.tpf = tpf

        # pull out the time
        self.time = Time(self.tpf.time, format='jd', scale=timescale)
        self.titlefordisplay = 'TIC{}\nCAM{} | ({},{}) | {:.0f}s'.format(
            tpf.tic_id, tpf.cam, tpf.col_cent, tpf.row_cent, tpf.cadence.to('s').value)

    def __getitem__(self, timestep):
        '''
        Return the image data for a given timestep.

        This function is called when you say `sequence[timestep]`.

        Parameters
        ----------
        timestep : int
                A timestep index (which element in the sequence do you want?)
        '''
        if timestep is None:
            return None
        else:
            return self.tpf.flux[timestep, :, :]

    def _gather_3d(self):
        '''
        Gather a 3D cube of images.
        '''
        return self.tpf.flux[:, :, :]



'''
Define a sequence of FITS images.
'''

from __future__ import print_function
from .filenameparsers import *
from .Image_Sequence import *

__all__ = ['FITS_Sequence']

class FITS_Sequence(Image_Sequence):
    '''
    A sequence of FITS images, with a time associated with each.
    '''

    def __init__(self, initial, ext_image=1, ext_primary=0, name='FITS',
                       use_headers=True,
                       use_filenames=False,
                       filenameparser=flexible_filenameparser,
                       timekey=None, timeformat=None,
                       **kwargs):
        '''
        Initialize a Sequence of FITS images. The goal is
        to create a list of FITS HDUs, one for each time.

        Parameters
        ----------

            initial : many possible types
                This is a flexible type of input, which could be...
                    - single FITS filename, and an extension to use.
                    - list of FITS filenames, and an extension to use.
                    - a glob-like search string containing '*'.
                    - single FITS HDUList, and an extension to use.
                    - list of loaded FITS HDULists, and an extension to use.

            ext_image : int
                The extension that is the image.

            ext_primary : int
                The extension that is the primary for the FITS.

            name : str
                The name of this sequence.

            use_headers : bool
                Should we extract (time) information from the headers?

            use_filenames : bool
                Should we extract (time) information from the filenames?

            filenameparser : function
                This function takes a filename as input, and returns a
                dictionary containing parsed keys and values derived from it.

            timekey : string
                Which temporal key should define the time axis?

            timeformt : string
                What's the time format for defining the time axis?

        '''

        # initialize the basic sequence
        Sequence.__init__(self, name=name)

        # we keep the HDUs out of memory, until we need them
        # (this should probably someday be rewritten as an iterator?)
        self._hdulists = None

        # ultimately, we want to make a list of filenames or HDULists
        if type(initial) == fits.HDUList:
            # if this is one HDUList, make it a list of them
            self._hdulists = [initial]
        elif type(initial) == str:
            # a search string
            if '*' in initial:
                self.filenames = np.sort(glob.glob(initial))
                #self.hdulists = [fits.open(f) for f in glob.glob(initial)]
            # a single file
            elif 'fit' in initial.lower():
                self.filenames = [initial]
                #self.hdulists = [fits.open(initial)]
        elif type(initial) == list:
            # a list of filenames
            if np.all([type(hdu) == fits.HDUList for hdu in initial]):
                self._hdulists = initial
            elif np.all([os.path.exists(s) for s in initial]):
                self.filenames = initial

        # if we're starting frmo hdulists, then get their filenames
        if self._hdulists is not None:
            self.filenames = [h.filename() for h in self._hdulists]
            self._hdulists = np.asarray(self._hdulists)

        self.filenames = np.asarray(self.filenames)

        # make sure this FITS_Sequence isn't empty
        #assert(len(self.filenames) > 0)

        self.ext_primary = ext_primary
        self.ext_image = ext_image
        if len(self.filenames) > 0:
        	self.ext_image = np.minimum(self.ext_image, len(self._get_hdulist(0))-1)

        self.temporal = {}
        self.static = {}
        self.spatial = {}

        # populate the temporal axes, somehow
        assert(use_headers or use_filenames)
        if use_headers:
            try:
                self._populate_from_headers()
            except:
                self.speak('unable to extract temporal things from headers for {}'.format(self))
        if use_filenames:
            try:
                self._populate_from_filenames(filenameparser=filenameparser)
            except:
                self.speak('unable to extract temporal things from filenames for {}'.format(self))

        # make sure a time axis gets defined
        self._define_time_axis(timekey=timekey, timeformat=timeformat)

        # make sure everything gets sorted by time
        self._sort()

    def _sort(self):
        '''
        Sort the images, and the temporals.
        '''

        # calculate sorting indices
        i = np.argsort(self.time.gps)

        # sort the temporal values
        for k in self.temporal.keys():
            self.temporal[k] = np.atleast_1d(self.temporal[k])[i]

        # sort the images
        self.filenames = self.filenames[i]
        if self._hdulists is not None:
            self._hdulists = self._hdulists[i]

        self.time = self.time[i]

    @property
    def N(self):
        '''
        How many elements are in this sequence?
        '''
        return len(self.filenames)

    def _get_hdulist(self, i):
        '''
        Return an HDUlist for the ith element in the sequence.

        (We might want to open them only one at a time, for memory's sake.)
        '''
        if self._hdulists is not None:
            return self._hdulists[i]
        else:
            return fits.open(self.filenames[i], memmap=False)

    def _clean_temporal(self):
        '''
        Move anything that's non-changing from temporal to static.
        '''
        # move non-changing things to static
        for k in list(self.temporal.keys()):
            if len(np.unique(self.temporal[k])) == 1:
                # if we have a one-element sequence, repeat it in both static and temporal
                if len(self.temporal[k]) == 1:
                    self.static[k] = self.temporal[k][0]
                else:
                    self.static[k] = self.temporal.pop(k)[0]
            else:
                self.temporal[k] = np.asarray(self.temporal[k])

        self.speak('the temporal keys for {} are {}'.format(self, list(self.temporal.keys())))
        self.speak('the static keys for {} are {}'.format(self, list(self.static.keys())))

    def _populate_from_filenames(self, filenameparser=flexible_filenameparser):
        '''
        Pull the basic information and temporal axis from the filenames.
        '''
        self.speak('populating {} information from the filenames (like {})'.format(self, self.filenames[0]))
        for i, f in enumerate(self.filenames):
            this = filenameparser(f)

            # create empty lists, if necessary
            if i == 0:
                for k in this.keys():
                    self.temporal[k] = []

            # tack this file onto the list
            for k in this.keys():
                self.temporal[k].append(this[k])

        # move static things away from temporal
        self._clean_temporal()

    def _populate_from_headers(self):
        '''
        Attempt to populate the sequence from the headers.
        '''

        self.speak('populating {} information from the headers'.format(self))

        # pull out the first HDUList in the sequence
        first = self._get_hdulist(0)

        # check to see if the requested image extension exists
        try:
            first[self.ext_image]
        except IndexError:
            self.speak('image extension {} not found, switching to 0'.format(self.ext_image))
            self.ext_image = 0

        # load the headers for the primary and the image
        pri, img = first[self.ext_primary].header, first[self.ext_image].header

        # if only a single image, everything is static (but can be viewed as temporal)
        if self.N == 1:
            for h in [pri, img]:
                for k in h.keys():
                    self.static[k] = h[k]
                    self.temporal[k] = [self.static[k]]
        else:

            # look through the unique extensions
            extensions = np.unique([self.ext_primary, self.ext_image])

            # create lists for each key in the headers
            for e in extensions:
                h = first[e].header
                for k in h.keys():
                    self.temporal[k] = []

            # compile all values from the headers
            for i in range(self.N):
                hdulist = self._get_hdulist(i)
                self.speak("populating header {} of {}".format(i+1, self.N), progress=True)
                for e in extensions:
                    h = hdulist[e].header
                    for k in h.keys():
                        self.temporal[k].append(h[k])

            # move static things away from temporal
            self._clean_temporal()

    def _define_time_axis(self, timekey=None, timeformat=None):

        # make up an imaginary GPS time (and keep track of whether it is fake)
        self.time = Time(np.arange(self.N), format='gps', scale='tdb')
        self._timeisfake = True

        # try to pull out a specific key
        try:
            assert(timekey is not None)
            self.time = Time(np.asarray(self.temporal[timekey]),
                        format=timeformat or 'gps',
                        scale=timescale)
            self.speak('using "{}" as the time axis'.format(timekey))
        except (AssertionError, KeyError):
            # try to pull a time axis from these
            for k in ['TIME', 'MJD', 'JD', 'BJD', 'BJD_TDB', 'DATE-OBS']:
                try:
                    # treat some value as a time
                    t = self.temporal[k]

                    # if it's already an astropy time, keep it as such
                    if isinstance(t, Time):
                        self.time = t
                    # make an astropy time out of the values
                    else:
                        self.time = Time(np.asarray(t),
                                    format=timeformat or guess_time_format(t),
                                    scale=timescale)
                        self._timeisfake = False
                    self.speak('guessing "{}" is good as the time axis'.format(k))
                    self._timeisfake = False
                except KeyError:
                    continue

    def __getitem__(self, timestep):
        '''
        Return the image data for a given timestep.

        This function is called when you say `sequence[timestep]`.

        Parameters
        ----------
        timestep : int
                A timestep index (which element in the sequence do you want?)
        '''
        if timestep is None:
            return None
        else:
            return self._get_hdulist(timestep)[self.ext_image].data



from .IllustrationBase import *
from ..frames import CameraFrame, cameras
from ..frames import CCDFrame, ccds
__all__ = ['CameraIllustration']


class CameraIllustration(IllustrationBase):
    '''
    An illustration for displaying a single Camera (unrotated, untransformed).
    '''
    illustrationtype = 'CameraIllustration'

    def __init__(self,
                 data=[],
                 orientation='horizontal',
                 sizeofcamera=4,
                 subplot_spec=None,
                 **framekw):
        '''
        Parameters
        ----------
        data : various
            Anything that can initialize a sequence with
            the make_image_sequence() helper function.

            or

            a dictionary with keys of 'ccd1', 'ccd2', 'ccd3', 'ccd4'
            that each contain something that can initialize a sequence

        orientation : str
            'horizontal' = camera 2 is right of camera 1
            'vertical'  = camera 2 is below camera 1

        sizeofcamera : float
            What's the size, in inches, to display a single camera?

        **framekw passed to CameraFrame
        '''


        # set up the basic geometry of the main axes
        rows, cols = 1, 1

        # what's the orientation of this illustration?
        self.orientation = orientation

        # craete the basic illustration
        IllustrationBase.__init__(self, rows, cols,
                                  figkw=dict(
                                      figsize=( sizeofcamera * cols,
                                                sizeofcamera * rows)),
                                  hspace=0.02, wspace=0.02,
                                  left=0.05, right=0.95,
                                  bottom=0.1, top=0.9,
                                  subplot_spec=subplot_spec)

        # populate the axes on the main camera grid
        ax = plt.subplot(self.grid[0, 0])


        # create a CameraFrame for this camera
        self.frames['camera'] = CameraFrame(illustration=self,
                                            ax=ax,
                                            data=make_image_sequence(data),
                                            **framekw)



from .IllustrationBase import *

class GenericIllustration(IllustrationBase):
    illustrationtype = 'Generic'

    def __init__(self,  imshows=[],
                        timeseries=[],
                        imshowheight=1.0,
                        imshowrows=1,
                        sizescale=3,
                        figsize=None,
                        dpi=None,
                        hspace=0.03,
                        wspace = 0.08,
                        left=0.1,
                        right=0.9,
                        bottom=0.2,
                        top=0.8,
                        **illkw):
        '''
        Initialize an illustration from list of frames.

        Parameters
        ----------

        imshows : list
            A list of imshow frames, to be displayed in a row on the top.

        timeseries : list
            A list of timeseries frames, to be displayed in rows below.

        imshowheight : float
            How tall is the imshow, compared to the entire timeseries block?

        sizescale : float
            An overall size scale, in inches.

        figsize : tuple
            Override the automated figure size with your own, in inches.

        **illkw keywords are passed to IllustrationBase
        '''

        # one row for imshows + one for each timeseries
        hasimshow = int(len(imshows) > 0)
        hastimeseries = int(len(timeseries) > 0)

        # one column for each imshow
        imshowcols = np.maximum(int(np.ceil(float(len(imshows))/imshowrows)), 1)
        rows = hasimshow * imshowrows + len(timeseries) + hastimeseries
        cols = np.maximum(1, imshowcols)

        # use the frame aspect ratios to set the width ratios
        width_ratios = [f.aspectratio for f in imshows][0:imshowcols]
        if len(width_ratios) == 0:
            width_ratios = [1]

        # set the height ratios, including a gap between imshows and timeseries
        height_ratios = (   hasimshow * imshowrows * [imshowheight] +
                            hastimeseries * [0.5] +
                            len(timeseries) * [1.0 / (1 + len(timeseries))])

        # set the sizes, trying to keep square imshows square
        wsize = sizescale * np.sum(width_ratios) * (1 + (cols - 1) * wspace) / (right - left)
        hsize = sizescale * np.sum(height_ratios) * \
            (1 + (rows - 1) * hspace) / (top - bottom)

        # allow the user to overwrite the automatic figure size
        figkw = dict(figsize=(figsize or (wsize, hsize)),
                     dpi=None or 100)

        # create a illustration to match this geometry
        IllustrationBase.__init__(self, rows, cols,
                                  figkw=figkw,
                                  hspace=hspace, wspace=wspace,
                                  left=left, right=right,
                                  bottom=bottom, top=top,
                                  height_ratios=height_ratios,
                                  width_ratios=width_ratios,
                                  **illkw)

        # add the imshows to the illustration
        for count, i in enumerate(imshows):
            row = count // imshowcols
            col = count % imshowcols

            # make sure to connect the frame back to this illustration
            i.illustration = self

            # create the axes in which this should sit
            i.ax = plt.subplot(self.grid[row, col])

            # register this in the frames dictionary
            try:
                n = i.name
                if n in self.frames:
                    n += '-{}'.format(col)
            except AttributeError:
                n = (row, col)
            self.frames[n] = i

        # add the timeseries to the illustration
        sharex = None  # you may want to change the shared axes for different visualizations
        for row, t in enumerate(timeseries):
            col = 0

            # make sure to connect the frame back to this illustration
            t.illustration = self
            # create the axes in which this should sit
            t.ax = plt.subplot(
                self.grid[row + hasimshow + 1, :], sharex=sharex)
            sharex = t.ax

            # turn off the axis labels
            if row != len(timeseries) - 1:
                plt.setp(t.ax.get_xticklabels(), visible=False)
            # register this in the frames dictionary
            try:
                n = t.name
                if n in self.frames:
                    n += '-{}'.format(row)
            except AttributeError:
                n = (row + hasimshow, col)
            self.frames[n] = t



from .FrameBase import *
from ..colors import cmap_norm_ticks
from ..sequences import make_image_sequence


class imshowFrame(FrameBase):
    '''
    An imshow frame can show a sequence of images, as an imshow.
    '''

    # what is the type of this frame?
    frametype = 'imshow'

    # what are the coordinate limits?
    xmin, xmax = None, None
    ymin, ymax = None, None

    def __init__(self,
                 name='image',
                 ax=None,
                 data=None,
                 title=None,
                 plotingredients=[  'image',
                                    'time',
                                    'colorbar',
                                    'arrows',
                                    'title'], # other options might be 'filename', 'axes', what else?
                 processingsteps=[],
                 firstframe=None,
                 cmapkw=dict(),
                 **kwargs):
        '''
        Initialize this imshowFrame, will can show a sequence of 2D images.

        Parameters
        ----------

        name : str
            A name to give this Frame.

        ax : matplotlib.axes.Axes instance
            All plotting will happen inside this ax.
            If set to None, the `self.ax attribute` will
            need to be set manually before plotting.

        data : Image_Sequence
            Any Sequence that contains 2D images.

        title : str
            A string to display as the title of this frame.
            If nothing, this will try to pull a title out
            of the data, or leave it blank.

        plotingredients : list
            A list of keywords indicating features that will be
            plotted in this frame. It can be modified either now
            when initializing the frame, or any time before
            calling `i.plot()` from the illustration.

        cmapkw : dict
            Dictionary of keywords to feed into the cmap generation.
        '''

        # initialize the frame base
        FrameBase.__init__(self, name=name,
                                 ax=ax,
                                 data=data,
                                 plotingredients=plotingredients,
                                 **kwargs)

        # ensure that the data are a sequence of images
        self.data = make_image_sequence(self.data, **kwargs)

        # if there's an image, use it to set the size
        try:
            self.xmin, self.ymin = 0, 0
            self.ymax, self.xmax = self.data[0].shape
        except (IndexError, AttributeError, TypeError):
            pass

        # try to figure out a title for the imshowFrame
        try:
            self.titlefordisplay = self.data.titlefordisplay
        except AttributeError:
            self.titlefordisplay = ''

        if title is not None:
            self.titlefordisplay = title

        # keep track of an extra keywords for generating the cmaps
        self.cmapkw = copy.copy(cmapkw) # why do I have to do this?

        # are there steps to apply to the image before displaying?
        self.processingsteps = processingsteps

        # should we plot something special for the first frame?
        self.firstframe = firstframe

    def _cmap_norm_ticks(self, *args, **cmapkw):
        '''
        Return the cmap and normalization.

        If the illustration has shared colorbar,
        then use the cmap and norm from there.

        Otherwise, make a colorbar for this frame.

        *args and **kwargs are passed to colors.cmap_norm_ticks
        '''

        if self.illustration.sharecolorbar:
            # pull the cmap and normalization from the illustration
            (self.plotted['cmap'],
             self.plotted['norm'],
             self.plotted['ticks']) = self.illustration._cmap_norm_ticks(**self.illustration.cmapkw)
            return (self.plotted['cmap'],
                    self.plotted['norm'],
                    self.plotted['ticks'])
        else:
            # use already-defined properties, or make new ones
            try:
                return  (self.plotted['cmap'],
                         self.plotted['norm'],
                         self.plotted['ticks'])
            except KeyError:

                # create the cmap from the given data
                (self.plotted['cmap'],
                 self.plotted['norm'],
                 self.plotted['ticks']) = cmap_norm_ticks(*args, **cmapkw)

                return  (self.plotted['cmap'],
                         self.plotted['norm'],
                         self.plotted['ticks'])

    def _ensure_colorbar_exists(self, image):
        '''
        Make sure this axes has its colorbar created.

        Parameters
        ----------

        image : the output returned from imshow

        '''
        # do we use a shared colorbar for the whole illustration?
        if self.illustration.sharecolorbar:
            self.speak('making sure a shared colorbar is set up for {}'.format(self))
            try:
                # if the illustration already has a colorbar, don't remake
                self.illustration.plotted['colorbar']
            except KeyError:
                # if the illustration needs a colorbar, make one!
                self.speak('added a shared colorbar for {}'.format(self.illustration))

                c = self.illustration._add_colorbar(image,
                                                    ax=None,
                                                    ticks=self.plotted['ticks'])
                self.illustration.plotted['colorbar'] = c
            return self.illustration.plotted['colorbar']
        # or do we just give this one frame its own colorbar?
        else:
            self.speak('making sure a unique colorbar is set up for {}'.format(self))
            try:

                self.plotted['colorbar']
            except KeyError:
                self.speak('added a unique colorbar for {}'.format(self))


                # create a colorbar for this illustration
                c = self.illustration._add_colorbar(image,
                                                    ax=self.ax,
                                                    ticks=self.plotted['ticks'])
                self.plotted['colorbar'] = c
            return self.plotted['colorbar']


    def draw_arrows(self, origin=(0, 0), ratio=0.05):
        '''
        Draw arrows on this Frame, to indicate
        the +x and +y directions.

        Parameters
        ----------
        origin : tuple
            The (x,y) coordinates of the corner of the arrows.
        ratio : float
            What fraction of the (longest) axis should the arrows span?
        '''


        # figure out the length of the arrows (in data units)
        try:
            xspan = np.abs(self.xmax - self.xmin)
            yspan = np.abs(self.ymax - self.ymin)
            length = ratio * np.maximum(xspan, yspan)
        except:
            length = 50

        # store the arrows in a dictionary
        arrows = {}

        # rotate into the display coordinates
        unrotatedx, unrotatedy = origin
        x, y = self._transformxy(*origin)
        arrow_kw = dict(zorder=10, color='black', width=length * 0.03, head_width=length *
                        0.3, head_length=length * 0.2, clip_on=False, length_includes_head=True)
        text_kw = dict(va='center', color='black', ha='center',
                       fontsize=7, fontweight='bold', clip_on=False)
        buffer = 1.4

        # +x arrow
        dx, dy = np.asarray(self._transformxy(unrotatedx + length, unrotatedy)) - \
            np.asarray(self._transformxy(unrotatedx, unrotatedy))
        arrows['xarrow'] = self.ax.arrow(x, y, dx, dy, **arrow_kw)
        xtextx, xtexty = self._transformxy(
            unrotatedx + length * buffer, unrotatedy)
        arrows['xarrowlabel'] = self.ax.text(xtextx, xtexty, 'x', **text_kw)

        # +y arrow
        dx, dy = np.asarray(self._transformxy(unrotatedx, unrotatedy + length)) - \
            np.asarray(self._transformxy(unrotatedx, unrotatedy))
        arrows['yarrow'] = self.ax.arrow(x, y, dx, dy, **arrow_kw)
        ytextx, ytexty = self._transformxy(
            unrotatedx, unrotatedy + length * buffer)
        arrows['yarrowlabel'] = self.ax.text(ytextx, ytexty, 'y', **text_kw)

        return arrows

    def plot(self, time=None):
        '''
        Generate the (initial) plot for this frame.

        Individual features can be modified afterwards,
        through the .ax (the plotted axes) or the .plotted
        (dictionary of plotted elements) attributes.

        Parameters
        ----------

        time : astropy Time
            The time to plot, defaults to the first with None.
        '''

        self.speak('plotting {} for the first time'.format(self))

        # make sure we point back at this frame
        plt.sca(self.ax)

        # kind of a kludge (to make the plots and cmaps reset)?
        self.plotted = {}

        # pull out the array to work on
        image, actual_time = self._get_image(time)

        # plot the image, as an imshow
        if ('image' in self.plotingredients):# and (image is not None):

            # pull out the cmap, normalization, and suggested ticks

            cmap, norm, ticks = self._cmap_norm_ticks(image, **self.cmapkw)

            # display the image for this frame
            extent = [0, image.shape[1], 0, image.shape[0]]

            # make a stacked image
            if self.firstframe is None:
                firstimage = image
            elif self.firstframe == 'median':
                #assert(np.size(image) < 10000 or self.data.N < 50)
                firstimage = self.data.median()

            self.plotted['image'] = self.ax.imshow(
                firstimage, extent=extent, interpolation='nearest', origin='lower', norm=norm, cmap=cmap)

            self.speak('added image of shape {} to {}'.format(firstimage.shape, self))

        # plot the colorbar
        if ('colorbar' in self.plotingredients) and 'image' in self.plotted:
            # add the colorbar
            self.plotted['colorbar'] = self._ensure_colorbar_exists(self.plotted['image'])

        # plot some text labeling the time
        if 'time' in self.plotingredients:
            if actual_time is None:
                timelabel = ''
            else:
                timelabel = self._timestring(actual_time)

            # add a time label
            self.plotted['time'] = self.ax.text(
                0.0, -0.02, timelabel, va='top', zorder=1e6, color='gray', transform=self.ax.transAxes)
            self.speak('added time label of "{}" on {}'.format(timelabel, self))

        # plot the arrows
        if 'arrows' in self.plotingredients:
            self.plotted['arrows'] = self.draw_arrows()
            self.speak('added arrows on {}'.format(self))

        # plot a title on this frame
        if 'title' in self.plotingredients:
            self.plotted['title'] = plt.title(self.titlefordisplay)
            self.speak('added title of "{}" to {}'.format(self.titlefordisplay, self))

        # plot lines and ticks for axes only if requested
        if 'axes' not in self.plotingredients:
            # turn the axes lines off
            plt.axis('off')

        # change the x and y limits, if need be
        self.ax.set_xlim(self.xmin, self.xmax)
        self.ax.set_ylim(self.ymin, self.ymax)
        self.ax.set_aspect('equal')

        # keep track of the current plotted timestep
        try:
            timestep = self._find_timestep(time)
        except:
            timestep = None
        self.currenttimestep = timestep

    def process_image(self, image):
        '''
        Apply any extra processing steps to the image
        (subract a median image, normalize, mask, ???)
        '''

        if 'subtractmedian' in self.processingsteps:
            processedimage = image - self.data.median()
            #self.speak('subtracted median image')
        elif 'subtractmean' in self.processingsteps:
            processedimage = image - self.data.mean()
        else:
            processedimage = image
        return processedimage

    def _get_image(self, time=None):
        '''
        Get the image at a given time (defaulting to the first time).
        '''

        try:
            if time is None:
                time = self._get_times()[0]
            timestep = self._find_timestep(time)
            rawimage = self.data[timestep]
            assert(rawimage is not None)

            processedimage = self.process_image(rawimage)
            image = self._transformimage(processedimage)
            actual_time = self._get_times()[timestep]
            # self.speak(" ")
            # self.speak(time, timestep)
        except (IndexError, AssertionError, ValueError):
            return None, None
        return image, actual_time

    def _get_alternate_time(self, time=None):
        '''
        The time are still a little kludgy.
        (Maybe this isn't even necessary?)
        '''
        for f in self.includes:
            try:
                timestep = f._find_timestep(time)
                actual_time = self._get_times()[timestep]
                break
            except IndexError:
                pass
        return actual_time

    def update(self, time):
        '''
        Update this frame to a particular time (for use in animations).
        '''
        # update the data, if we need to
        timestep = self._find_timestep(time)
        image, actual_time = self._get_image(time)
        if image is None:
            return

        if timestep != self.currenttimestep:
            if 'image' in self.plotingredients:
                self.plotted['image'].set_data(image)
            if 'time' in self.plotingredients:
                self.plotted['time'].set_text(self._timestring(actual_time))
        self.currenttimestep = timestep



from .GenericIllustration import *
from ..frames import CameraFrame
from ..frames import ZoomFrame

__all__ = ['SingleCameraWithZoomIllustration']


class SingleCameraWithZoomIllustration(GenericIllustration):
    '''
    For displaying a single Camera.
    '''
    illustrationtype = 'SingleCameraWithZoomIllustration'

    def __init__(self, data=[], zoomposition=(0, 0), zoomsize=(10, 10), **kwargs):
        '''
        **kwargs are passed to both make_image_sequence and the frames
        '''

        # create a CameraFrame for this camera
        i_camera = CameraFrame( illustration=self,
                                data=make_image_sequence(data, **kwargs),
                                **kwargs)

        # add the ZoomFrame, connected to the camera
        i_zoom = ZoomFrame(     illustration=self,
                                source=i_camera,
                                position=zoomposition,
                                size=zoomsize,
                                **kwargs)

        GenericIllustration.__init__(self, imshows=[i_camera, i_zoom])



'''
Wrappers to define some default visualizations.
'''

from __future__ import print_function

from .imports import *
from .zoom import *
from .illustrations import *
from .postage.stamps import *
from .sequences import *

def camera_from_filename(f):
    '''
    Try to guess the camera from the filename.

    Parameters
    ----------

    filename : str
        The name of a fits file, with 'cam?' in it somewhere.
    '''

    try:
        return os.path.basename(f).split('cam')[1][0]
    except IndexError:
        return '?'

def organize_sequences(pattern='*.fits',
                       filenameparser=flexible_filenameparser,
                       ext_image=1, use_headers=False, use_filenames=True,
                       timekey='cadence'):
    '''
    Take a group of filenames, and group them in
    one of the following ways:

    1) a dictionary (with camera as key)
       of lists of filenames
    2) a dictionary (with camera as key)
       of dictionaries (with ccd as key)
       of lists of filenames

    Parameters
    ----------
    pattern : str, list
        If a string, a file-search pattern (e.g. with '*') for glob
        If a list, a list of filenames.

    filenameparser : function
        This is a function that takes in a filename,
        and returns a dictionary that contains (maybe)
        a "camera" keyword with an integer camera number
        and a "ccd" keyword with an integer ccd number.
        It is used to decide how to group images.

     ext_image=0, use_headers=False, use_filenames=True,
    '''

    # create a list of filenames
    if type(pattern) == list:
        # if given a list, use those as the filenames
        filenames = pattern
    else:
        # if given a string, use it as a glob search string
        filenames = list(np.sort(glob.glob(pattern)))

    # create an empty dictionary of cameras + CCDs
    cameras = {'cam{}'.format(c):{
                    'ccd{}'.format(d):[] for d in [1,2,3,4,'?']}
               for c in [1,2,3,4,'?']}

    # loop through images, and figure out where they belong
    for f in filenames:

        # parse the filename into a dictionary
        i = filenameparser(f)

        # figure out an appropriate camera key
        if 'camera' in i:
            camerakey = 'cam{}'.format(i['camera'])
        else:
            camerakey = 'cam?'

        # figure out an appropriate CCD key
        if 'ccd' in i:
            ccdkey = 'ccd{}'.format(i['ccd'])
        else:
            ccdkey = 'ccd?'

        # store the filename into its appropriate key
        cameras[camerakey][ccdkey].append(f)

    # convert each list of filenames into a sequence
    sequences = {}
    for camerakey in cameras.keys():
        sequences[camerakey] = {}
        for ccdkey in cameras[camerakey].keys():
            files = cameras[camerakey][ccdkey]
            if len(files) != 0:
                sequences[camerakey][ccdkey] = make_image_sequence(files,
                                                   ext_image=ext_image,
                                                   use_headers=use_headers,
                                                   use_filenames=use_filenames,
                                                   timekey=timekey)


    # if there aren't multiple CCDs, compress each camera to single list
    organized = {}
    for camerakey in sequences.keys():
        keys = list(sequences[camerakey].keys())
        if len(keys) == 1:
            organized[camerakey] = sequences[camerakey][keys[0]]
        elif len(keys) > 1:
            organized[camerakey] = sequences[camerakey]
    print('yo!')
    # return the grouped filenames
    return organized


def illustratefits( pattern='*.fits',
                    zoomposition=None, zoomsize=(10,10),
                    filenameparser=flexible_filenameparser,
                    **illustrationkw):
    '''
    Make an Illustration from a group of FITS files.

    If one (or fewer) camera is found (by parsing the filename),
    then a CameraIllustration is returned.

    If multiple cameras are found (by parsing the filename),
    then a FourCameraIllustration is returned.

    If multiple cameras and multiple CCDs are found (by parsing the filename),
    then a FourCameraOfCCDsIllustration is returned.

    Parameters
    ----------
    pattern : string, list
        If a string, a file-search pattern (e.g. with '*') for glob
        If a list, a list of filenames.

    zoomposition : tuple
        (x,y) position of where a zoom should be placed

    zoomsize : tuple
        (x,y) size of the zoom window to create

    **illustrationkw : dict
        Keyword arguments will be passed to the Illustration
    '''

    # make a "group" of filenames (either a list, a dictionary, a dictionary of dictionares)
    data = organize_sequences(pattern, filenameparser=filenameparser)
    assert(type(data) == dict)

    # is there only a single camera represented?
    singlecamera = len(data) == 1

    def merge_inputs(kw):
        # a small tool to merge dictionaries of inputs
        # (because Python 2 can't handle `**data,**inputs` in functions)
        inputs = dict(**kw)
        for k in illustrationkw.keys():
            inputs[k] = illustrationkw[k]
        return inputs

    # display these data, one way or another
    if singlecamera:
        # (there's only one camera we're trying to display)

        # pull out the (only) sequence
        cam = list(data.keys())[0]
        if type(data[cam]) is dict:
            # (one camera, and it has multiple CCDs)
            illustration = CameraOfCCDsIllustration(**merge_inputs(data[cam]))
        else:
            # (one camera, with one pseudo-CCD)
            seq = data[cam]
            if zoomposition is not None:
                # display the single image, with a zoom box added
                illustration = SingleCameraWithZoomIllustration(seq, zoomposition=zoomposition, zoomsize=zoomsize, **illustrationkw)
            else:
                # display the single image, just by itself
                illustration = CameraIllustration(seq, **illustrationkw)
            # the zoom option works only for single-frame illustrations (FIXME?)
    elif len(data) > 1:
        # (there are multiple cameras to display)

        # figure out if there's one or more CCDs
        multipleccds = False
        for k in data.keys():
            multipleccds = multipleccds or (type(data[k]) is dict)

        if multipleccds:
            # (multiple cameras, and each is multiple CCDs)
            illustration = FourCameraOfCCDsIllustration(**merge_inputs(data))
        else:
            # (multiple cameras, and each is one psuedo-CCD)
            illustration = FourCameraIllustration(**merge_inputs(data))


    return illustration

def illustratestamps(pattern='stamps/spm*/*.npy', get_camera=camera_from_filename, zoom=50, **kw):
    '''
    FIXME -- I'm not positive this still works.

    Make an Illustration from a group of Stamps.

    If one (or fewer) cameras are found through
    the 'cam{}' substring, a CameraIllustration illustration is returned.

    If more than on cameras are found through
    the 'cam{}' substring, a FourCameraIllustration illustration is returned.

    Parameters
    ----------
    pattern : string, list
        If a string, a file-search pattern (e.g. with '*') for glob
        If a list, a list of filenames.

    get_camera : function
        Returns 1,2,3,4 or "?" based on strings in the filename.

    **kw : dict
        Keyword arguments will be passed to the Illustration

    '''

    # pull out the files
    files = glob.glob(pattern)
    data = {'cam{}'.format(c): [] for c in [1, 2, 3, 4, '?']}

    # sort them into cameras
    for f in files:
        cam = get_camera(f)
        data['cam{}'.format(cam)].append(f)

    # remove the empty cameras
    keys = list(data.keys())
    for k in keys:
        if len(data[k]) == 0:
            data.pop(k)

    singlecamera = len(data) == 1
    # figure out how to display them
    if singlecamera:
        cam = 'camera'
        illustration = CameraIllustration(**kw)
    elif len(data) > 1:
        cam = None
        illustration = FourCameraIllustration(**kw)
    else:
        return

    for k in data.keys():
        stamps = [Stamp(f) for f in data[k]]
        for s in stamps:
            if singlecamera:
                camera = 'camera'
            else:
                camera = 'cam{}'.format(s.cam)
            frame = add_stamp(illustration, s, zoom=zoom, camera=camera)
    return illustration
